class ExoscaleAPIException(Exception):
    """
    Base exception for exoscale API errors.
    Base for all other exception classes, it allows to catch all exoscale exceptions with a single except block.
    """

    def __init__(self, message, response=None):
        super().__init__(message)
        self.response = response


class ExoscaleAPIClientException(ExoscaleAPIException):
    """
    For client-side errors (4xx).
    Shows that the client sent a bad request.
    """

    pass


class ExoscaleAPIServerException(ExoscaleAPIException):
    """
    For server-side errors (5xx).
    Shows the server encountered an error while processing the request.
    """

    pass


class ExoscaleAPIAuthException(ExoscaleAPIException):
    """
    For authentication-related errors (403).
    Shows the server encountered an error while processing the request.
    """

    pass



"""
Exoscale Partner API client

This module provides a client for the Exoscale Partner API, which allows
distributors to manage sub-organizations.
"""

import json
from pathlib import Path

from .generator import create_client_class
from .v2 import Client as V2Client


with open(Path(__file__).parent.parent / "partner-api.json", "r") as f:
    partner_api_spec = json.load(f)
    BasePartnerClient = create_client_class(partner_api_spec)


class Client(BasePartnerClient):
    """
    Partner API client with Exoscale authentication.

    This client provides access to distributor operations for managing
    sub-organizations. It uses the same authentication mechanism as the
    V2 API client.

    Args:
        key (str): Exoscale API key
        secret (str): Exoscale API secret
        url (str): Override endpoint URL (optional)
        zone (str): Exoscale zone (optional)

    Example:
        >>> from exoscale.api.partner import Client
        >>> client = Client("EXO...", "secret")
        >>> orgs = client.list_distributor_organizations()
    """

    def __init__(self, key, secret, *args, url=None, **kwargs):
        # Initialize with Partner API endpoint
        partner_url = (
            url if url else "https://partner-api.exoscale.com/v1.alpha"
        )
        super().__init__(*args, url=partner_url, **kwargs)

        # Reuse the v2 client's authentication mechanism
        v2_client = V2Client(key, secret, *args, url=url, **kwargs)

        self.http_client = v2_client.http_client
        self.key = key

        self._v2_client = v2_client

    def __repr__(self):
        return (
            f"<Client endpoint={self.endpoint} "
            f"key={self.key} secret=**masked**>"
        )



"""

``exoscale.api.v2`` provides a low-level client targeting Exoscale's
OpenAPI-based V2 API. This client is dynamically generated from the OpenAPI
definition shipped with the package.

Examples:
    Creating a client targeting one of the available zones:

    >>> from exoscale.api.v2 import Client
    >>> c = Client("api-key", "api-secret", zone="de-fra-1")
    >>> c.list_instances()
    {'instances': []}

    Creating a client targeting an endpoint specifically:

    >>> from exoscale.api.v2 import Client
    >>> c = Client("api-key", "api-secret", url="https://api-ch-gva-2.exoscale.com/v2")
    >>> c.list_instances()
    {'instances': []}

    Waiting for an asynchronous operation to complete:

    >>> from exoscale.api.v2 import Client
    >>> c = Client("api-key", "api-secret", zone="ch-gva-2")
    >>> version = c.list_sks_cluster_versions()["sks-cluster-versions"][0]
    >>> operation = c.create_sks_cluster(
    ...     cni="cilium",
    ...     name="my-cluster",
    ...     level="starter",
    ...     version=version,
    ... )
    >>> c.wait(operation["id"])
    {'id': 'e2047130-b86e-11ef-83b3-0d8312b2c2d7',
     'state': 'success',
     'reference': {
         'id': '8561ee34-09f0-42da-a765-abde807f944b',
         'link': '/v2/sks-cluster/8561ee34-09f0-42da-a765-abde807f944b',
         'command': 'get-sks-cluster'}}

    In case of a conflict between argument names and Python keywords, ``**kwargs`` syntax can be used:

    >>> from exoscale.api.v2 import Client
    >>> c = Client("api-key", "api-secret", zone="ch-gva-2")
    >>> c.list_events(**{"from": "2025-03-01"})
    [{'handler': 'authenticate', 'source-ip': 'x.x.x.x', 'message': 'User user@exoscale.com: authenticate', 'status': 200, 'timestamp': '2025-03-10T14:52:34Z'}, {'handler': 'create session', 'source-ip': 'x.x.x.x', 'message': 'User user@exoscale.com: create session', 'status': 200, 'timestamp': '2025-03-10T14:52:46Z'}]
"""

import json
import time
from pathlib import Path

import requests
from exoscale_auth import ExoscaleV2Auth

from .generator import (
    _return_docstring,
    create_client_class,
    ExoscaleAPIClientException,
    ExoscaleAPIServerException,
)


def _poll_interval(run_time):
    """
    Returns the wait interval before next poll, given the current run time of a job.
    We poll
     - every 3 seconds for the first 30 seconds
     - then increase linearly to reach 1 minute at 15 minutes of run time
     - then every minute
    """
    # y = a * x + b. Solve a and b for:
    # 60 = a * 900 + b
    # 3 = a * 30 + b
    a = 57 / 870
    b = 3 - 30 * a
    min_wait = 3
    max_wait = 60
    interval = a * run_time + b
    interval = max(min_wait, interval)
    interval = min(max_wait, interval)
    return interval


def _time():
    return time.time()


def _sleep(start_time):
    run_time = _time() - start_time
    interval = _poll_interval(run_time)
    return time.sleep(interval)


with open(Path(__file__).parent.parent / "openapi.json", "r") as f:
    api_spec = json.load(f)
    BaseClient = create_client_class(api_spec)


class Client(BaseClient):
    def __init__(self, key, secret, *args, url=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.WAIT_ABORT_ERRORS_COUNT = 5

        client = requests.Session()
        client.auth = ExoscaleV2Auth(key, secret)
        self.http_client = client
        self.key = key

    def __repr__(self):
        return (
            f"<Client endpoint={self.endpoint}"
            f" key={self.key} secret=***masked***>"
        )

    def wait(self, operation_id: str, max_wait_time: int = None):
        """
        Wait for completion of an asynchronous operation.

        Args:
            operation_id (str)
            max_wait_time (int): When set, stop waiting after this time in
              seconds. Defaults to ``None``, which waits until operation
              completion.

        Returns:
            {ret}
        """
        start_time = _time()
        subsequent_errors = 0
        while True:
            try:
                result = self.get_operation(id=operation_id)
                subsequent_errors = 0
            except ExoscaleAPIServerException as e:
                subsequent_errors += 1
                if subsequent_errors >= self.WAIT_ABORT_ERRORS_COUNT:
                    raise ExoscaleAPIServerException(
                        "Server error while polling operation"
                    ) from e
                _sleep(start_time)
                continue
            state = result["state"]
            if state == "success":
                return result
            elif state in {"failure", "timeout"}:
                raise ExoscaleAPIServerException(
                    f"Operation error: {state}, {result['reason']}"
                )
            elif state == "pending":
                run_time = _time() - start_time
                if max_wait_time is not None and run_time > max_wait_time:
                    raise ExoscaleAPIClientException(
                        "Operation max wait time reached"
                    )
                _sleep(start_time)
            else:
                raise ExoscaleAPIServerException(
                    f"Invalid operation state: {state}"
                )


Client.wait.__doc__ = Client.wait.__doc__.format(
    ret=_return_docstring(
        Client._api_spec, Client._by_operation["get-operation"]["operation"]
    )
)



import copy
from itertools import chain

import requests

from .exceptions import (
    ExoscaleAPIAuthException,
    ExoscaleAPIClientException,
    ExoscaleAPIServerException,
)


def _get_in(payload, keys):
    """
    Returns the value in a nested dict, where items is a sequence of keys.
    """
    k, *ks = keys
    ret = payload[k]
    if ks:
        return _get_in(ret, ks)
    else:
        return ret


def _get_ref(api_spec, path):
    root, *parts = path.split("/")
    if root != "#":
        raise AssertionError("Non-root path start", root, path)

    # We're going to mutate payload later on, so make a full copy to avoid
    # altering api_spec.
    payload = copy.deepcopy(_get_in(api_spec, parts))

    for name, desc in payload.get("properties", {}).items():
        if "$ref" in desc:
            resolved_schema = _get_ref(api_spec, desc["$ref"])
            for k, v in desc.items():
                if k != "$ref":
                    resolved_schema[k] = v
            payload["properties"][name] = resolved_schema
    payload["$schema"] = "http://json-schema.org/draft-04/schema"
    return payload


_type_translations = {
    "string": "str",
    "integer": "int",
    "object": "dict",
    "array": "list",
    "boolean": "bool",
    "number": "float",
}


def _status_code_docstring(api_spec, operation, status_code):
    [ctype] = operation["responses"][status_code]["content"].keys()
    return_schema = operation["responses"][status_code]["content"][ctype][
        "schema"
    ]
    if "$ref" in return_schema:
        ref = _get_ref(api_spec, return_schema["$ref"])
        if (
            "properties" in ref
            and ref["type"] == "object"
            and "description" in ref
        ):
            body = {}
            for name, prop in ref["properties"].items():
                if "$ref" in prop:
                    _ref = _get_ref(api_spec, prop["$ref"])
                    item = _ref
                else:
                    item = prop
                typ = _type_translations[item["type"]]
                desc = prop.get("description")
                if "enum" in item:
                    choices = "``, ``".join(map(repr, item["enum"]))
                    desc += f". Values are ``{choices}``"
                suffix = f": {desc}" if desc else ""
                normalized_name = name.replace("-", "_")
                body[normalized_name] = (
                    f"**{normalized_name}** ({typ}){suffix}."
                )

            doc = (
                f"dict: {ref['description']}. A dictionary with the following keys:"
                + "\n\n          * ".join([""] + list(body.values()))
            )
        elif "description" in ref:
            doc = f"{_type_translations[ref['type']]}: {ref['description']}."
        else:
            doc = _type_translations[ref["type"]]
    else:
        doc = _type_translations[return_schema["type"]]
    return doc


def _return_docstring(api_spec, operation):
    status_codes_docs = [
        "{status_code}: {ret_type}".format(
            status_code=status_code,
            ret_type=_status_code_docstring(api_spec, operation, status_code),
        )
        for status_code in operation["responses"].keys()
    ]

    return "\n\n        ".join(status_codes_docs)


class BaseClient:
    _api_spec = None
    _by_operation = None

    def __init__(self, url=None, **kwargs):
        if url is None:
            server = self._api_spec["servers"][0]
            variables = {
                var_name: var["default"]
                for var_name, var in server["variables"].items()
            }
            for k, v in kwargs.items():
                if k not in server["variables"]:
                    raise TypeError(f"Unhandled keyword argument {k!r}.")
                if choices := server["variables"][k].get("enum"):
                    if v not in choices:
                        choices_repr = "', '".join(choices)
                        raise TypeError(
                            f"Invalid {k}: must be one of '{choices_repr}'."
                        )
                variables[k] = v

            self.endpoint = server["url"].format(**variables)
        else:
            self.endpoint = url

        self.http_client = requests.Session()

    def __repr__(self):
        return f"<Client endpoint={self.endpoint}>"

    def _call_operation(self, operation_id, parameters=None, body=None):
        op = self._by_operation[operation_id]

        path = op["path"]
        query_params = {}
        path_params = {}
        if parameters is None:
            parameters = {}
        for param in op["operation"].get("parameters", []):
            name = param["name"]
            if param["required"] and name not in parameters:
                raise ValueError(f"Missing mandatory param {name!r}")
            if name in parameters:
                value = parameters[name]
                if param["in"] == "path":
                    path_params[name] = value
                elif param["in"] == "query":
                    query_params[name] = value

        path = path.format(**path_params)

        url = f"{self.endpoint}{path}"

        json = {}
        if body is not None:
            # TODO validate
            json["json"] = body

        response = self.http_client.request(
            method=op["verb"].upper(), url=url, params=query_params, **json
        )

        # Error handling
        if response.status_code == 403:
            raise ExoscaleAPIAuthException(
                f"Authentication error {response.status_code}: {response.text}",
                response,
            )
        if 400 <= response.status_code < 500:
            raise ExoscaleAPIClientException(
                f"Client error {response.status_code}: {response.text}",
                response,
            )
        elif response.status_code >= 500:
            raise ExoscaleAPIServerException(
                f"Server error {response.status_code}: {response.text}",
                response,
            )

        return response.json()


def _args_docstring(parameters, body):
    return "\n\n        ".join(chain(parameters.values(), body.values()))


def _create_operation_call(
    py_operation_name, operation_name, operation, api_spec
):
    docstring = """{summary}

    Args:
        {args}

    Returns:
        {ret}
    """

    parameters = {}
    body = {}
    normalized_names = {}
    for param in operation.get("parameters", []):
        name = param["name"]
        if "$ref" in param["schema"]:
            ref = _get_ref(api_spec, param["schema"]["$ref"])
            typ = _type_translations[ref["type"]]
        else:
            typ = _type_translations[param["schema"]["type"]]
        normalized_name = name.replace("-", "_")
        normalized_names[normalized_name] = name
        parameters[normalized_name] = f"{normalized_name} ({typ})."

    if "requestBody" in operation:
        schema = operation["requestBody"]["content"]["application/json"][
            "schema"
        ]
        if "$ref" in schema:
            ref = _get_ref(api_spec, schema["$ref"])
            properties = ref["properties"]
        else:
            properties = schema["properties"]

        for name, prop in properties.items():
            if "$ref" in prop:
                ref = _get_ref(api_spec, prop["$ref"])
                item = ref
            else:
                item = prop
            typ = _type_translations[item["type"]]
            desc = prop.get("description")
            if "enum" in item:
                choices = "``, ``".join(map(repr, item["enum"]))
                desc += f". Must be one of ``{choices}``"
            suffix = f": {desc}" if desc else ""
            normalized_name = name.replace("-", "_")
            normalized_names[normalized_name] = name
            body[normalized_name] = f"{normalized_name} ({typ}){suffix}."

    def _api_call(self, *args, **kwargs):
        if args:
            raise TypeError(
                f"{py_operation_name}() only accepts keyword arguments."
            )
        _params = {}
        _body = {}
        for k, v in kwargs.items():
            if k not in normalized_names:
                raise TypeError(f"Unhandled keyword argument {k!r}.")
            api_name = normalized_names[k]
            if k in parameters:
                _params[api_name] = v
            elif k in body:
                _body[api_name] = v
            else:
                raise TypeError(f"Unhandled keyword argument {k!r}.")

        if not _body:
            _body = None

        return self._call_operation(
            operation_name, parameters=_params, body=_body
        )

    _api_call.__name__ = py_operation_name
    _api_call.__qualname__ = f"Client.{py_operation_name}"
    _api_call.__doc__ = docstring.format(
        summary=operation["summary"],
        args=_args_docstring(parameters, body),
        ret=_return_docstring(api_spec, operation),
    )
    return _api_call


def _client_docstring(api_spec):
    template = """Create an API client.

    Args:
        key (str): Exoscale API key.

        secret (str): Exoscale API secret.

        url (str): Endpoint URL to use. Defaults to ``{default_server!r}``.

        {dynamic_args}

    Returns:
        Client: A configured API client.
    """
    servers = []
    args = {}
    for server in api_spec["servers"]:
        servers.append(server["url"])
        for name, variable in server["variables"].items():
            if name in args:
                continue
            choices = "``, ``".join(map(repr, variable["enum"]))
            typ = type(variable["default"]).__name__
            args[name] = (
                f"{name} ({typ}): one of ``{choices}``."
                f" Defaults to ``{variable['default']!r}``."
            )
    dynamic_args = "\n\n        ".join(args.values())
    return template.format(
        servers="``, ``".join(map(repr, servers)),
        default_server=servers[0],
        dynamic_args=dynamic_args,
    )


def create_client_class(api_spec):
    by_operation = {}
    for path, item in api_spec["paths"].items():
        for verb, operation in item.items():
            if verb not in {
                "get",
                "put",
                "post",
                "delete",
                "head",
                "options",
                "patch",
                "trace",
            }:
                raise AssertionError(
                    "Unhandled path item object (https://swagger.io/specification/#pathItemObject) field",  # noqa
                    verb,
                )
            by_operation[operation["operationId"]] = {
                "verb": verb,
                "path": path,
                "operation": operation,
            }

    class_name = "Client"
    bases = [BaseClient]
    class_attributes = {
        "_api_spec": api_spec,
        "_by_operation": by_operation,
        "__doc__": _client_docstring(api_spec),
    }

    for operation_name, operation in by_operation.items():
        py_operation_name = operation_name.replace("-", "_")
        op_fn = _create_operation_call(
            py_operation_name,
            operation_name,
            operation["operation"],
            api_spec,
        )

        class_attributes[py_operation_name] = op_fn

    cls = type(class_name, tuple(bases), class_attributes)
    return cls
